import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import AxesGrid
from mpl_toolkits.mplot3d import Axes3D
import numpy as np


def show_overview(d, cell_to_preview, gain_to_preview, out_folder=None, infix=None):

    for module, data in d.items():
        fig = plt.figure(figsize=(20,20))
        grid = AxesGrid(fig, 111,
                        nrows_ncols=(2, 2),
                        axes_pad=(0.9, 0.15),
                        label_mode="1",
                        share_all=True,
                        cbar_location="right",
                        cbar_mode="each",
                        cbar_size="7%",
                        cbar_pad="2%",
                        )
        i = 0
        for key, item in data.items():
            cf = 0
            if "ThresholdsDark" in key:
                cf = -1
            if len(item.shape) == 4:
                med = np.nanmedian(item[...,cell_to_preview, gain_to_preview + cf])
            else:
                med = np.nanmedian(item[...,cell_to_preview])
                
            bound = 0.2
            while(np.count_nonzero((item[...,cell_to_preview, gain_to_preview + cf] < med-np.abs(bound*med)) |
                                   (item[...,cell_to_preview, gain_to_preview + cf] > med+np.abs(bound*med)))/item[...,cell_to_preview, gain_to_preview + cf].size > 0.01):            
                bound *=2
            
            if "BadPixelsDark" in key:
                im = grid[i].imshow(np.log2(item[...,cell_to_preview, gain_to_preview + cf]), interpolation="nearest",
                                    vmin=0, vmax=8, aspect='auto')
            else:
                if len(item.shape) == 4:
                    im = grid[i].imshow(item[...,cell_to_preview, gain_to_preview + cf], interpolation="nearest",
                                       vmin=med-np.abs(bound*med), vmax=np.abs(med+bound*med), aspect='auto')
                else:
                    im = grid[i].imshow(item[...,cell_to_preview], interpolation="nearest",
                                       vmin=med-np.abs(bound*med), vmax=med+np.abs(bound*med), aspect='auto')
            cb = grid.cbar_axes[i].colorbar(im)
            cb.set_label_text("ADU" if key != "BadPixels" else "Bad Pixel Code")

            grid[i].text(5, 20, key, color="w" if key != "BadPixels" else "k", fontsize=20)

            i += 1
        grid[0].text(5, 50, module, color="r" if key != "BadPixels" else "k", fontsize=20)
        if out_folder and infix:
            fig.savefig("{}/dark_analysis_{}_module_{}.png".format(out_folder,
                                                                   infix,
                                                                   module))


def rebin(a, *args):
    '''rebin ndarray data into a smaller ndarray of the same rank whose dimensions
    are factors of the original dimensions. eg. An array with 6 columns and 4 rows
    can be reduced to have 6,3,2 or 1 columns and 4,2 or 1 rows.
    example usages:
    https://scipy-cookbook.readthedocs.io/items/Rebinning.html
    >>> a=rand(6,4); b=rebin(a,3,2)
    >>> a=rand(6); b=rebin(a,2)
    '''
    shape = a.shape
    lenShape = len(shape)
    factor = np.asarray(shape)//np.asarray(args)
    evList = ['a.reshape('] + \
             ['args[%d],factor[%d],'%(i,i) for i in range(lenShape)] + \
             [')'] + ['.sum(%d)'%(i+1) for i in range(lenShape)] + \
             ['/factor[%d]'%i for i in range(lenShape-1)]

    ta = eval(''.join(evList))
    return ta.astype(np.uint32), np.indices([s + 1 for s in ta.shape])


def plot_badpix_3d(data, definitions, title=None, rebin_fac=2, azim=22.5):
    od = data
    d, dims = rebin(od.astype(np.uint32), od.shape[0]//rebin_fac, od.shape[1]//rebin_fac, od.shape[2])
    xx, yy, zz = dims
    voxels = d.astype(np.bool)
    colors = np.full(voxels.shape, '#FFFFFF')
    cols = definitions

    for k, c in cols.items():
        colors[d == k] = c[1]

    fig = plt.figure(figsize=(15,10))
    ax = fig.gca(projection="3d")
    ax.voxels(xx*rebin_fac, yy*rebin_fac, zz, voxels, facecolors=colors)
    ax.view_init(elev=25., azim=azim)
    ax.set_xlabel("pixels")
    ax.set_ylabel("pixels")
    ax.set_zlabel("memory cell")
    ax.set_xlim(0, np.max(xx)*rebin_fac)
    ax.set_ylim(0, np.max(yy)*rebin_fac)
    ax.set_zlim(0, np.max(zz))

    for k, c in cols.items():
        ax.plot([-1,], [-1,], color=c[1], label=c[0])
    ax.legend()
    if title:
        ax.set_title(title)

def create_constant_overview(constant, name, cells, vmin=None, vmax=None,
                             entries=3, out_folder=None, infix=None,
                             badpixels=None, gmap=None, marker=None):
    """
    Create a step plot for constant data across memory cells for requested
    gain entries

    :param constant: dict with constants for each module.
    :param name: Name to be used for the x-axis
    :param cells: Number of memory cells
    :param vmin: plot minumim value boundaries
    :param vmax: plot maximum value boundaries
    :param entries: (int)number of gain entries.
        A tuple specifying the range can also be used.
    TODO: remove unused inputs from notebooks.
    :param out_folder: out_folder for showing .md table statistics
    :param infix: infix for the output png image
    :param badpixels: A list of 2 elements.
        badpixels[0] has the dict with badpixels constant for each module
        and badpixels[1] has the value to apply for bad pixels.
    :param gmap: A list with len equal to number of gain entires.
        if not supported, a default would be used for 3 entries.
        ['High gain', 'Medium gain', 'Low gain']
    :param marker: A list of line markers for each gain entry.
        default: ['']*entries
    :return:
    """
    if gmap is None:
        gmap = ['High gain', 'Medium gain', 'Low gain']
    if marker is None:
        marker = ['']*entries
    fig = plt.figure(figsize=(10, 5))
    ax = fig.add_subplot(111)
    for g in range(entries):
        table = []
        dbp = None

        for qm in constant.keys():
            if len(constant[qm].shape) == 4:
                d = constant[qm][..., g]
                if badpixels is not None and isinstance(badpixels, list):
                    dbp = np.copy(d)
                    dbp[badpixels[0][qm][..., g] > 0] = badpixels[1]

            else:
                # e.g. DSSC
                d = constant[qm]
                if badpixels is not None and isinstance(badpixels, list):
                    dbp = np.copy(d)
                    dbp[badpixels[0][qm] > 0] = badpixels[1]
            # TODO: check if not used to remove.
            table.append([name, qm, gmap[g], np.nanmean(d), np.nanmedian(d),
                          np.nanstd(d)])

            ax.step(np.arange(cells), np.nanmean(d, axis=(0,1)),
                    label=gmap[g], color=f'C{g}', marker=marker[g])
            # Plotting good pixels only if bad-pixels were given
            if dbp is not None:
                ax.step(np.arange(cells), np.nanmean(dbp, axis=(0, 1)),
                        label=f'Good pixels {gmap[g]}', color=f'C{g}',
                        linestyle='--')
    ax.set_xlabel("Memory cell")
    ax.set_ylabel(name)
    ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    ax.set_title(f"{name} Median per Cell".format(gmap[g]))
    if vmin and vmax:
        ax.set_ylim(vmin, vmax)
